#ifndef _vteGridForVtk_h
#define _vteGridForVtk_h

#include "vteFrame.h"
#include "vteStorageClass.h"

#include "vtkPoints.h"
#include "vtkStructuredGrid.h"
#include "vtkStructuredGridGeometryFilter.h"
#include "vtkStructuredGridOutlineFilter.h"
#include "vtkAppendPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkLODActor.h"
#include "vtkLookupTable.h"

class vteGridFrame;
class vteGridForVtk;

/*******************************************
vteGridFrameCollection

GridFrameCollection is a collection of grid frames,
each containing volume data that can be visualized
using isosurfaces or cutting planes

This class also defines the overlying topology
of the grid, which stays constant across the frames
********************************************/
class vteGridFrameCollection:public vteFrameCollection
{
public:
	vteGridFrameCollection();
	~vteGridFrameCollection();

	vteGridFrame *AddFrame();
	vteGridFrame *GetFrame(long index = -1);
	unsigned int MembersInFrame (int index = -1) {return Nn;}

	/*end of interface*/
	int AddVar(char *name);	
	void SetActiveVar(unsigned int c) {active_var=c;}
	unsigned int GetActiveScalar() {return active_var;}
	unsigned int GetNumberOfVars() {return n_vars;}
	char *GetVarName() {return var_name[active_var];}
	char *GetVarName(int index) {return var_name[index];}
	
	void SetNumberOfNodes(unsigned int x, unsigned int y, unsigned z) {Nx=x;Ny=y;Nz=z;Nn=Nx*Ny*Nz;}
	void SetSpacing(double x, double y, double z) {dx=x;dy=y;dz=z;}
	void SetOrigin(double x, double y, double z) {x0=x;y0=y;z0=z;}
	
	vteGridData* GetGridData(int index=-1);

	inline unsigned int GetNx() {return Nx;}
	inline unsigned int GetNy() {return Ny;}
	inline unsigned int GetNz() {return Nz;}
	long GetNumberOfNodes() {return Nn;}
	
	inline double GetX0() {return x0;}
	inline double GetY0() {return y0;}
	inline double GetZ0() {return z0;}

	inline double GetDx() {return dx;}
	inline double GetDy() {return dy;}
	inline double GetDz() {return dz;}

	vteRange *GetAbsoluteRange(int component=-1); 
	vteRange *GetUserRange(int component=-1); 
	void SetUserRange(double min, double max); 

	vteGridForVtk *GetGridForVtk();

protected:
	int Nx, Ny, Nz;
	long Nn;
	double dx, dy, dz;
	double x0, y0, z0;
	
	int n_vars;
	char *(var_name[50]);
	int active_var;

private:
	vteRange *absolute_range;
	vteRange *user_range;
	vteGridFrame **frame;

	/*VTK data structures*/
	vteGridForVtk *grid_for_vtk;
};
  
/*****************************************
vteGlyphFrame

GlyphFrame contains info about all glyphs 
existing in a particular frame. 
******************************************/
class vteGridFrame:public vteCollection
{
public:
	vteGridFrame (int nn, int nv);	/*number of nodes and vars, this needs to match the collection!*/
	~vteGridFrame ();

	int SetVarBlock (vteGridData *data);
	vteGridData *GetVarBlock(int var) {return &data[var*nn];}

private:
	vteGridData *data;
	int nn, nv;
};

/****************************************
vteGridForVtk

VTK representation of grid data
*****************************************/
class vteGridForVtk:public vteStorageClass
{
public:
	static vteGridForVtk* vteGridForVtk::New(vteGridFrameCollection *input);
	void Delete();

	vteGridForVtk();
	~vteGridForVtk();

	vteRange *GetAbsoluteRange(int component=-1) {return input->GetAbsoluteRange(component);}
	vteRange *GetUserRange(int component=-1) {return input->GetUserRange(component);}
	vteGridFrameCollection *GetInput() {return input;}
	vtkLookupTable *GetLut() {return lut;}
	vtkStructuredGrid *GetStructuredGrid() {return grid;}
	vtkLODActor *GetActor() {return actor;}
	double *GetPtrToScalars() {return ptr_to_scalars;}
	vteGridForVtk *UpdateGrid(vteGridFrameCollection *frame);

protected:
	vtkStructuredGrid *grid;
	vtkStructuredGridGeometryFilter *geometry_filter;
	vtkStructuredGridOutlineFilter *outline_filter;

	vtkAppendPolyData *filter;
	vtkPolyDataMapper *mapper;
	vtkLODActor *actor;
	double *ptr_to_scalars;

	vtkLookupTable *lut;
	vteGridFrameCollection *input;
};

#endif
