#include <vtkConeSource.h>
#include <vtkActor.h>
#include <vtkCamera.h>

#include "QVTKWidget.h"

#include "vteGUIWindow.h"
#include "vteFileReader.h"
#include "vteGeometry.h"

//***************************************
//***************************************
vteGUIWindow::vteGUIWindow( QMdiArea* workspace, vteGUI *gui, QStatusBar *status_bar, 
						   const char* file_name, Qt::WindowFlags flags): QMdiSubWindow(workspace, flags)
{   
   /*init variables*/
	network=NULL;
   
/*TODO: 2010 LB: THIS SHOULD BE MOVED OUT OF THE CONSTRUCTOR*/

   /*create VTK widget which takes care of the drawing*/
   qvtk_window=vteVisualizationWindow::New(workspace,file_name);
   if (qvtk_window==NULL) close();
   qvtk_window->SetStatusBar(status_bar);

   if (qvtk_window->Execute()==vteERROR) 
	{
		close();
		status_bar->showMessage(tr("Loading operation failed!"));
		return;
   }
  
   qvtk_widget = qvtk_window->GetWidget();
   setWidget(qvtk_widget);

  
   playback_rate=3;
   playback_timer = new QTimer(this);
   connect( playback_timer, SIGNAL(timeout()), this, SLOT(PlaybackTimerOverhead()) );
   playing=false;

   record_name_index=0;
   recording=false;
   record_timer=new QTimer(this);
   connect(record_timer, SIGNAL(timeout()), this, SLOT(RecordTimerOverhead()));

   network_connected=false;

   iso_properties_auto=true;
   parent_gui=gui;
   this->status_bar=status_bar;

   workspace->addSubWindow(this);
   //qvtk_window->show();
}

vteGUIWindow::~vteGUIWindow()
{
	//if (playback_timer!=NULL) delete playback_timer;
	if (qvtk_window!=NULL) qvtk_window->Delete();
	if (network!=NULL) delete network;
}

//PlaybackTimerOverhead - updates to next frame
void vteGUIWindow::PlaybackTimerOverhead()
{
	qvtk_window->UpdateToFrame(qvtk_window->GetCurrentFrame()+1);
}

int vteGUIWindow::TogglePlay()
{
	if (playing)
	{
		playback_timer->stop();
		playing=false;
	}
	else
	{
		playback_timer->start((int)(1000.0/playback_rate),FALSE);
		playing=true;
	}
 
	return vteOK;
}

void vteGUIWindow::SetPlaybackRate(int rate)
{
	playback_rate=rate;
	if (playing) playback_timer->changeInterval((int)(1000.0/playback_rate));
}

//starts recording snapshots at a preset interval
int vteGUIWindow::ToggleRecord()
{
	if (recording)
	{
		record_timer->stop();
		recording=false;
	}
	else
	{
		record_timer->start(500,FALSE);
		recording=true;
	}
 
	return vteOK;
}

void vteGUIWindow::RecordTimerOverhead()
{
	strstream name;
	name<<"capVTE_recording_";
	if (record_name_index<1000)
	{
		name<<"0";
		if (record_name_index<100)
		{
			name<<"0";
			if (record_name_index<10)	name<<"0";
		}
	}

	name<<record_name_index<<".iv"<<'\0';
	qvtk_window->WriteJPEG(name.str(),100);    //no compression
	record_name_index++;
}

/*------------------------------------NETWORK-----------------------------------
  The following functions handle network related activity
*/

/*StartNetwork
*/
int vteGUIWindow::StartNetwork(long port, const QString &name, const QString &avatar, bool server, const QString &address)
{
	/* create dialog box for displaying network messages*/
	network_status = new vteNetworkStatus(this);
	
	/* start the network*/
	network = new vteNetwork(this, port, name, -1, server, address);
	if (network==NULL) ErrorMessage("Cannot start network");

	number_of_clients=0;

	return vteOK;
}

void vteGUIWindow::StopNetwork()
{
	network->Disconnect();
	network_connected=false;
}

/*NetworkConnectionStatus
This function is called with the status of OUR connection
*/
void vteGUIWindow::NetworkConnectionStatus(int type)
{
	switch (type)
	{
		case ACCEPT: network_connected=true;
			         status_bar->message("connection accepted");
				     break;
		case DATA_MISMATCH: status_bar->message("data mismatch");
			         network_connected=true;
				     break;
		case REJECT: network_connected=false;
					status_bar->message("connection refused");
				    break;
	}

	if ((type==ACCEPT)||(type==DATA_MISMATCH))
    {
		//avatar_collection=vteGeometryForVtkCollection::New();
		//if (avatar_collection==NULL) ErrorMessage("insufficient memory for the network");

		connect(network->GetClient(),SIGNAL(NewData(QString &)),this,SLOT(ProcessNetworkData(QString &)));
		network_timer = new QTimer(this);
		connect(network_timer, SIGNAL(timeout()), this, SLOT(NetworkTimer()));
		network_timer->start(250,FALSE); //lets sample 4x a second
    }
}

/*ProcessNetworkData
Take an action based on the string received from the net
*/
void vteGUIWindow::ProcessNetworkData(QString &data)
{
	char command[300];
	int sender_id;
	int index;
	QTextIStream line(&data);
	vteGUIWindow *active_window;

	line>>command; //this will contain "VTE"
	line>>sender_id; //this is the id of the sender socket

	line>>command;

	active_window=(vteGUIWindow*)parent_gui->GetActiveWindow();
	parent_gui->SetActiveWindow(this);
	network_connected=false;  //let's lie to the toggle action

	if (stricmp(command,"NEW_CONNECTION")==0)
	{
		network_client_info[number_of_clients].id=sender_id;
		line>>command;   //name
		network_client_info[number_of_clients].name=command;

		vtkVectorText *source=vtkVectorText::New();
		source->SetText(command);

		vtkPolyDataMapper *mapper=vtkPolyDataMapper::New();
		mapper->SetInput(source->GetOutput());

		vtkFollower *follower=vtkFollower::New();
		follower->SetMapper(mapper);
		follower->SetCamera(qvtk_window->GetRenderer()->GetActiveCamera());
		follower->GetProperty()->SetColor(0,0,0);
		follower->SetScale(2.5);
		follower->SetOrigin(-5,0,0);

		qvtk_window->GetRenderer()->AddActor(follower);

		network_client_info[number_of_clients].source=source;
		network_client_info[number_of_clients].mapper=mapper;
		network_client_info[number_of_clients].follower=follower;
		number_of_clients++;
	}
	else if (stricmp(command,"TOGGLE_GEOMETRY")==0)
		parent_gui->ToggleGeometry();
	else if (stricmp(command,"TOGGLE_GLYPHS")==0)
		parent_gui->ToggleGlyphs();
	else if (stricmp(command,"TOGGLE_GRID")==0)
		parent_gui->ToggleGrid();
	else if (stricmp(command,"TOGGLE_ISOSURFACE")==0)
		parent_gui->ToggleIsosurface();
	else if (stricmp(command,"TOGGLE_PLANES")==0)
		parent_gui->TogglePlanes();
	else if (stricmp(command,"TOGGLE_PLAY")==0)
		parent_gui->TogglePlay();
	else if (stricmp(command,"TOGGLE_SCALAR_GRADIENT")==0)
		parent_gui->ToggleScalarGradient();
	else if (stricmp(command,"AVATAR")==0)
	{
		for (index=0;index<number_of_clients;index++)
		if (network_client_info[index].id==sender_id)
		{
			line>>network_client_info[index].avatar;
			AddAvatar(network_client_info[index].avatar);
		}
	}  
	else if (stricmp(command,"CAMERA")==0)
	{
		double x,y,z,i,j,k;
		line>>x>>y>>z>>i>>j>>k;

		//lets see who sent this
		for (index=0;index<number_of_clients;index++)
		if (network_client_info[index].id==sender_id)
		{
			vtkActor *actor=avatar_collection->GetGeometry(i)->GetGeometryForVtk()->GetActor();
			actor->SetPosition(x,y,z);
			actor->SetOrientation(-i,-j,-k);
			actor->AddOrientation(0,180,0);
			network_client_info[index].follower->SetPosition(x,y,z);
			qvtk_window->GetRenderer()->Render();
			break;
		}
	}     
	else cout<<"unknown network command: "<<command<<endl;

	data.clear();
}

/*AddAvatar
*/
void vteGUIWindow::AddAvatar(const char *name)
{
	vteFileReader *file_reader=vteFileReader::New(name);
	if (file_reader->Execute()!=vteERROR)
	{
		//generate a new entry in the avatar geometry table
		if (file_reader->GetGeometryCollection()->GetNumberOfMembers()>0)
		{
			/*LB 2010*/
			//avatar_collection->NewGeometryForVtk(file_reader->GetGeometryCollection()->GetGeometry(0),NULL);
			vtkActor *actor=avatar_collection->GetGeometry(avatar_collection->GetNumberOfMembers()-1)->GetGeometryForVtk()->GetActor();
			actor->SetScale(2.5);
			qvtk_window->GetRenderer()->AddActor(actor);
		}
	}

 file_reader->Delete();
}

/*NetworkTimer
 Collect info about the camera and send it off
*/
void vteGUIWindow::NetworkTimer()
{
	vtkCamera *camera=qvtk_window->GetRenderer()->GetActiveCamera();
	double x,y,z;
	double *v;

	QString message;
	QTextOStream stream(&message);

	camera->GetPosition(x,y,z);
	v=camera->GetOrientation();
	stream<<"CAMERA "<<x<<" "<<y<<" "<<z<<" "<<v[0]<<" "<<v[1]<<" "<<v[2];
	network->PropagateMessage(message);
}
