#include "vteGeometry.h"
#include "vtkPoints.h"
#include "vtkPointData.h"
#include "vtkDoubleArray.h"

vteGeometry::vteGeometry()
{
	nodes=NULL;
	cells=NULL;
	colors=NULL;
	geometry_for_vtk=NULL;

	number_of_nodes=0;
	number_of_cells=0;
	number_of_colors=0;
}

vteGeometry::~vteGeometry()
{
	if (nodes!=NULL) delete nodes;
	if (cells!=NULL) delete cells;
	if (colors!=NULL) delete colors;
}

int vteGeometry::AddCells(vteCell *new_cells,int nc)
{
	if (!cells)
		cells = new vteCell[nc];
	else
		cells = (vteCell*) realloc (cells, (number_of_cells+nc)*sizeof(vteCell));

	for (int i=0;i<nc;i++)
		cells[number_of_cells+i] = new_cells[i];

	number_of_cells += nc;
	return vteOK;
}

int vteGeometry::AddNodes(vteNode *new_nodes, int nn)
{
	if (!nodes)
		nodes = new vteNode[nn];
	else
		nodes = (vteNode*) realloc (nodes, (number_of_nodes+nn)*sizeof(vteNode));

	for (int i=0;i<nn;i++)
		nodes[number_of_nodes+i] = new_nodes[i];

	number_of_nodes += nn;
	return vteOK;
}

int vteGeometry::AddColor(vteColor color)
{
	if (!colors)
		colors = new vteColor;
	else
		colors = (vteColor*) realloc (colors, (number_of_colors+1)*sizeof(vteColor));
	
	colors[number_of_colors] = color;
	number_of_colors++;
	return vteOK;
}

vteGeometryForVtk* vteGeometry::GetGeometryForVtk()
{
	if (!geometry_for_vtk)
		geometry_for_vtk = vteGeometryForVtk::New(this,NULL);

	return geometry_for_vtk;
}

//vteGeometryCollection
vteGeometryCollection::vteGeometryCollection()
{
	geometry = NULL;
}

vteGeometryCollection::~vteGeometryCollection()
{
	for (int i=0;i<members;i++) delete geometry[i];
}

vteGeometry *vteGeometryCollection::NewGeometry()
{
	if (geometry)
		geometry = (vteGeometry**) realloc(geometry, sizeof(vteGeometry*)*(members+1));
	else
		geometry = (vteGeometry**) new (vteGeometry**);

	/*define collection upon NewGeometry*/
	Define(this);
	
	geometry[members] = new vteGeometry;
	members++;
	return geometry[members-1];
}

vteGeometry *vteGeometryCollection::GetGeometry(int i)
{
	if ((i>=0) && (i<members)) return geometry[i];
	return NULL;
}



vteGeometryForVtk::vteGeometryForVtk()
{
	polydata=NULL;
	mapper=NULL;
	lut=NULL;
	actor=NULL;
	ptr_to_nodes=NULL;
	ptr_to_scalars=NULL;
	color_scalars_backup=NULL;
	box_index=NULL;
	input=NULL;
}

vteGeometryForVtk::~vteGeometryForVtk()
{
	if (polydata!=NULL) polydata->Delete();
	if (mapper!=NULL) mapper->Delete();
	if (lut!=NULL) lut->Delete();
	if (actor!=NULL) actor->Delete();
	if (color_scalars_backup!=NULL) delete color_scalars_backup;
	if (box_index!=NULL) delete box_index;
}

/******************************
vteGeometryForVtk::New

Constructs VTK data structures for representing geometry
*******************************/
vteGeometryForVtk* vteGeometryForVtk::New(vteGeometry *geometry, vteGridFrameCollection *grid_frame_collection)
{
	vteGeometryForVtk *geometry_for_vtk = new vteGeometryForVtk;
	if (geometry_for_vtk==NULL) return NULL;

	unsigned int i;
	double rgba[4];

	geometry_for_vtk->input=geometry;

	if (IsDefined(geometry)==false) return NULL;

	//generate vtkPolyData
	vtkPoints *points=vtkPoints::New();
	vtkDoubleArray *scalars=vtkDoubleArray::New();

	unsigned int non=geometry->GetNumberOfNodes();
	geometry_for_vtk->polydata = vtkPolyData::New();

	if (!points || !scalars || !geometry) return NULL;

	points->SetNumberOfPoints(non);
	scalars->SetNumberOfValues(non);

	vteNode *node = geometry->GetNodes();
	
	for (i=0;i<non;i++)
	{
		points->SetPoint(i, node[i].x);
		scalars->SetValue(i, node[i].color_index);
	}

	geometry_for_vtk->polydata->Allocate(geometry->GetNumberOfCells());
	geometry_for_vtk->polydata->SetPoints(points);
	geometry_for_vtk->polydata->GetPointData()->SetScalars(scalars);

	points->Delete();
	scalars->Delete();

	//assign pointers for fast memory update
	geometry_for_vtk->ptr_to_nodes = geometry_for_vtk->polydata->GetPoint(0);
	geometry_for_vtk->ptr_to_scalars = (double*)geometry_for_vtk->polydata->GetPointData()->GetScalars()->GetVoidPointer(0);

	geometry_for_vtk->color_scalars_backup=new double[non];
	if (geometry_for_vtk->color_scalars_backup==NULL) return NULL;
	memcpy(geometry_for_vtk->color_scalars_backup,geometry_for_vtk->ptr_to_scalars,non*sizeof(float));

	//now that we have the node points, define cells
	vtkIdType node3[3];
	vteCell *cell = geometry->GetCells();

	for (i=0;i<geometry->GetNumberOfCells();i++)
	{
		node3[0]=cell[i].node1;
		node3[1]=cell[i].node2;
		node3[2]=cell[i].node3;
		geometry_for_vtk->polydata->InsertNextCell(VTK_TRIANGLE,3,node3);
	}

	//generate look up table to map user defined colors
	geometry_for_vtk->lut=vtkLookupTable::New();
	if (geometry_for_vtk->lut==NULL) return NULL;

	geometry_for_vtk->lut->SetNumberOfTableValues(geometry->GetNumberOfColors());
	vteColor *color = geometry->GetColors();

	for (i=0;i<geometry->GetNumberOfColors();i++)
	{
		rgba[0]=color[i].r;
		rgba[1]=color[i].g;
		rgba[2]=color[i].b;
		rgba[3]=color[i].a;
		geometry_for_vtk->lut->SetTableValue(i,rgba);
	}
	geometry_for_vtk->lut->SetTableRange(0,geometry->GetNumberOfColors()-1);

	//generate geometry mapper
	geometry_for_vtk->mapper=vtkPolyDataMapper::New();
	if (geometry_for_vtk->mapper==NULL) return NULL;

	geometry_for_vtk->mapper->SetInput(geometry_for_vtk->polydata);
	geometry_for_vtk->mapper->SetLookupTable(geometry_for_vtk->lut);
	geometry_for_vtk->mapper->SetScalarRange(geometry_for_vtk->lut->GetRange());

	geometry_for_vtk->actor=vtkLODActor::New();
	if (geometry_for_vtk->actor==NULL) return NULL;
	geometry_for_vtk->actor->SetMapper(geometry_for_vtk->mapper);

	if (0 && IsDefined(grid_frame_collection)) //generate box_index indexes
	{
		double dx,dy,dz;
		unsigned int nx,ny,nz;
		double ox,oy,oz;
		unsigned int n_i,n_j,n_k;

		dx = grid_frame_collection->GetDx();
		dy = grid_frame_collection->GetDy();
		dz = grid_frame_collection->GetDz();
		nx = grid_frame_collection->GetNx();
		ny = grid_frame_collection->GetNy();
		nz = grid_frame_collection->GetNz();
		ox = grid_frame_collection->GetX0();
		oy = grid_frame_collection->GetY0();
		oz = grid_frame_collection->GetZ0();

		geometry_for_vtk->box_index=new vteBoxIndex[non];
		vteBoxIndex *box_index=geometry_for_vtk->box_index;

		vteNode *node=geometry->GetNodes();

		for (i=0;i<non;i++)
  		{
			if ((node[i].x[0]<ox) || (node[i].x[1]<oy) || (node[i].x[2]<oz)
			||(node[i].x[0]>=(ox+(nx-1)*dx)) || (node[i].x[1]>=(oy+(ny-1)*dy)) || (node[i].x[2]>=(oz+(nz-1)*dz)))
					  box_index[i].valid=false;
			else
			{
					box_index[i].valid=true;
					n_i=(unsigned int)((node[i].x[0]-ox)/dx);
					n_j=(unsigned int)((node[i].x[1]-oy)/dy);
					n_k=(unsigned int)((node[i].x[2]-oz)/dz);
					box_index[i].A=n_k*ny*nx+n_j*nx+n_i;
					box_index[i].B=n_k*ny*nx+n_j*nx+(n_i+1);
					box_index[i].C=n_k*ny*nx+(n_j+1)*nx+n_i;
					box_index[i].D=n_k*ny*nx+(n_j+1)*nx+(n_i+1);
					box_index[i].E=(n_k+1)*ny*nx+n_j*nx+n_i;
					box_index[i].F=(n_k+1)*ny*nx+n_j*nx+(n_i+1);
					box_index[i].G=(n_k+1)*ny*nx+(n_j+1)*nx+n_i;
					box_index[i].H=(n_k+1)*ny*nx+(n_j+1)*nx+(n_i+1);
			}
		}
	}
 
	Define(geometry_for_vtk);
	return geometry_for_vtk;
}

void vteGeometryForVtk::Delete()
{
	delete this;
}
