#include "vteGrid.h"
#include "vtkPointData.h"
#include "vtkProperty.h"
#include "vtkDoubleArray.h"


vteGridData* vteGridFrameCollection::GetGridData(int index)
{
	if (index<0) index=frame_index;
	return frame[index]->GetVarBlock(active_var);
}

/************************
vteGridFrameCollection::vteGridFrameCollection

Constructor
*************************/
vteGridFrameCollection::vteGridFrameCollection()
{
	Nx=Ny=Nz=0;
	dx=dy=dz=0;
	x0=y0=z0=0;
	n_vars=0;
	var_name[0] = NULL;
	active_var = 0;
	frame = NULL;
	grid_for_vtk = NULL;
	
	absolute_range=new vteRange[50];
	user_range=new vteRange[50];
}
/********************
vteGridFrameCollection::~vteGridFrameCollection

Destructor
*********************/
vteGridFrameCollection::~vteGridFrameCollection()
{
	if (absolute_range!=NULL) delete absolute_range;
	if (user_range!=NULL) delete user_range;

	for (int i=0;i<n_vars;i++)
		delete var_name[i];
}

/*******************************
vteGridFrameCollection::GetAbsoluteRange

Return range for the specified variable
********************************/
vteRange *vteGridFrameCollection::GetAbsoluteRange(int var)
{
	if (var<0) return &absolute_range[active_var];
	else return &absolute_range[var];
}

vteRange *vteGridFrameCollection::GetUserRange(int var)
{
	if (var<0) return &user_range[active_var];
	else return &user_range[var];
}

void vteGridFrameCollection::SetUserRange(double min, double max)
{
	user_range[active_var].min=min;
	user_range[active_var].max=max;

	vtkLookupTable *lut = grid_for_vtk->GetLut();
	lut->SetTableRange(min, max);
	lut->Build();
}
  
int vteGridFrameCollection::AddVar(char *name)
{
	var_name[n_vars]=new char[50];
	if (var_name[n_vars]==NULL)
		return ErrorMessage("AddComponent - insufficient memory");

	/*should do some length checking here...*/
	strcpy(var_name[n_vars],name);
	n_vars++;

	SetTupleSize(sizeof(vteGridData)*n_vars);
	return vteOK;
}

vteGridFrame::vteGridFrame (int nn, int nv)
{
	data = new vteGridData[nn*nv];
	this->nn = nn;
	this->nv = nv;
}

vteGridFrame::~vteGridFrame ()
{
	delete data;
}

vteGridFrame* vteGridFrameCollection::AddFrame()
{
	if (!frame)
		frame = (vteGridFrame**) new vteGridFrame*;
	else
		frame = (vteGridFrame**) realloc(frame, sizeof(vteGridFrame*)*(members+1));

	vteGridFrame *new_frame = new vteGridFrame(Nn, n_vars);
	frame[members++] = new_frame;
	max_frame = members-1;		/*also increment max_frame*/

	return new_frame;
}

vteGridForVtk *vteGridFrameCollection::GetGridForVtk()
{
	if (!grid_for_vtk)
		grid_for_vtk = vteGridForVtk::New(this);

	return grid_for_vtk;
}

/*****************************
vteGridFrame::SetVarBlock

Sets scalar data for all variables
******************************/
int vteGridFrame::SetVarBlock (vteGridData *data)
{
	memcpy(this->data,data,sizeof(vteGridData)*nn*nv);
	return vteOK;
}

/***************************
vteGridFroVtk::vteGridForVtk

Constructor
****************************/
vteGridForVtk::vteGridForVtk()
{
	grid=NULL;
	geometry_filter=NULL;
	outline_filter=NULL;
	filter=NULL;
	mapper=NULL;
	actor=NULL;
	ptr_to_scalars=NULL;
	lut=NULL;
	input=NULL;
}

vteGridForVtk::~vteGridForVtk()
{
	if (grid!=NULL) grid->Delete();
	if (geometry_filter!=NULL) geometry_filter->Delete();
	if (outline_filter!=NULL) outline_filter->Delete();
	if (filter!=NULL) filter->Delete();
	if (mapper!=NULL) mapper->Delete();
	if (actor!=NULL) actor->Delete();
	if (lut!=NULL) lut->Delete();
}

vteGridForVtk* vteGridForVtk::New(vteGridFrameCollection *input)
{
	vteGridData *frame_data;
	unsigned int non,noc; //number of nodes, components
	unsigned int i,j,k;
	unsigned int index;
	double x,y,z;

	if (!IsDefined(input))
		return NULL;
	
	vteGridForVtk *grid_for_vtk=new vteGridForVtk;
	if (grid_for_vtk==NULL) return NULL;
  
	grid_for_vtk->input=input;

	non=input->GetNumberOfNodes();
	noc=input->GetNumberOfVars();

	vtkPoints *nodes=vtkPoints::New();
	if (nodes==NULL) return NULL;
	nodes->SetNumberOfPoints(non);

	vtkDoubleArray *scalars=vtkDoubleArray::New();
	if (scalars==NULL) return NULL;
	scalars->SetNumberOfValues(non);

	frame_data=input->GetGridData(0);

	//cout<<"active_component= "<<input->GetActiveComponent()<<endl;

	index=0;

	z=input->GetZ0();
	for (k=0;k<input->GetNz();k++)
	{
		y=input->GetY0();

		for (j=0;j<input->GetNy();j++)
		{
			x=input->GetX0();
			for (i=0;i<input->GetNx();i++)
			{
				nodes->SetPoint(index,x,y,z);
				x+=input->GetDx();
				index++;
			}
			y+=input->GetDy();
		}
		z+=input->GetDz();
	}
	index=0;

	for (i=0;i<non;i++)
	{
		scalars->SetValue(index,frame_data[index]);
		index++;
	}

	grid_for_vtk->grid=vtkStructuredGrid::New();
	if (grid_for_vtk->grid==NULL) return NULL;

	grid_for_vtk->grid->SetDimensions(input->GetNx(),input->GetNy(),input->GetNz());
	grid_for_vtk->grid->SetPoints(nodes);
	nodes->Delete();
	grid_for_vtk->grid->GetPointData()->SetScalars(scalars);

	grid_for_vtk->ptr_to_scalars=(double *)grid_for_vtk->grid->GetPointData()->GetScalars()->GetVoidPointer(0);
	scalars->Delete();

	grid_for_vtk->outline_filter=vtkStructuredGridOutlineFilter::New();
	if (grid_for_vtk->outline_filter==NULL) return NULL;
	grid_for_vtk->outline_filter->SetInput(grid_for_vtk->grid);

	grid_for_vtk->geometry_filter=vtkStructuredGridGeometryFilter::New();
	if (grid_for_vtk->geometry_filter==NULL) return NULL;
	grid_for_vtk->geometry_filter->SetExtent(0,input->GetNx(),0,input->GetNy(),0,input->GetNz());
	grid_for_vtk->geometry_filter->SetInput(grid_for_vtk->grid);

	grid_for_vtk->filter=vtkAppendPolyData::New();
	if (grid_for_vtk->filter==NULL) return NULL;
	grid_for_vtk->filter->AddInput(grid_for_vtk->geometry_filter->GetOutput());
	grid_for_vtk->filter->AddInput(grid_for_vtk->outline_filter->GetOutput());

	grid_for_vtk->mapper=vtkPolyDataMapper::New();
	if (grid_for_vtk->mapper==NULL) return NULL;
	grid_for_vtk->mapper->SetInput(grid_for_vtk->filter->GetOutput());

	grid_for_vtk->actor=vtkLODActor::New();
	if (grid_for_vtk->actor==NULL) return NULL;
	grid_for_vtk->actor->SetMapper(grid_for_vtk->mapper);
	grid_for_vtk->actor->GetProperty()->SetColor(1,0,1);

	grid_for_vtk->lut=vtkLookupTable::New();
	if (grid_for_vtk->lut==NULL) return NULL;
	grid_for_vtk->lut->SetHueRange(0.667,0);
	grid_for_vtk->lut->SetNumberOfColors(255);
	grid_for_vtk->lut->SetTableRange(grid_for_vtk->GetUserRange()->min, grid_for_vtk->GetUserRange()->max);
	grid_for_vtk->lut->Build();

	Define(grid_for_vtk);

	return grid_for_vtk;
}

void vteGridForVtk::Delete()
{
	delete this;
}

vteGridForVtk* vteGridForVtk::UpdateGrid(vteGridFrameCollection *input)
{
	vteGridData *frame_data;
	unsigned int non,noc; //number of nodes, components
	unsigned int i,j,k;
	unsigned int index;
	float x,y,z;

	if (IsDefined(input)==false)
		return NULL;
	
	vteGridForVtk *grid_for_vtk=new vteGridForVtk;
	if (grid_for_vtk==NULL) return NULL;

	/*!!!!!!!!!! THIS IS NOT RIGHT, SHOULD PASS IN FRAME INDEX INSTEAD*/
	grid_for_vtk->input=input;

	non=input->GetNx()*input->GetNy()*input->GetNz();
	noc=input->GetNumberOfVars();

	vtkPoints *nodes=vtkPoints::New();
	if (nodes==NULL) return NULL;
	nodes->SetNumberOfPoints(non);

	vtkDoubleArray *scalars=vtkDoubleArray::New();
	if (scalars==NULL) return NULL;
	scalars->SetNumberOfValues(non);

	frame_data=input->GetGridData();

	index=0;
	z=input->GetZ0();
	for (k=0;k<input->GetNz();k++)
	{
		y=input->GetY0();

		for (j=0;j<input->GetNy();j++)
		{
			x=input->GetX0();
			for (i=0;i<input->GetNx();i++)
			{
				nodes->SetPoint(index,x,y,z);
				x+=input->GetDx();
				index++;
			}
			y+=input->GetDy();
		}
		z+=input->GetDz();
	}
	index=0;

	for (i=0;i<non;i++)
	{
		scalars->SetValue(index,frame_data[index]);
    	index++;
	}

	grid_for_vtk->grid=vtkStructuredGrid::New();
	if (grid_for_vtk->grid==NULL) return NULL;

	grid_for_vtk->grid->SetDimensions(input->GetNx(),input->GetNy(),input->GetNz());
	grid_for_vtk->grid->SetPoints(nodes);
	nodes->Delete();
	grid_for_vtk->grid->GetPointData()->SetScalars(scalars);

	grid_for_vtk->ptr_to_scalars=(double *)grid_for_vtk->grid->GetPointData()->GetScalars()->GetVoidPointer(0);
	scalars->Delete();

	grid_for_vtk->outline_filter=vtkStructuredGridOutlineFilter::New();
	if (grid_for_vtk->outline_filter==NULL) return NULL;
	grid_for_vtk->outline_filter->SetInput(grid_for_vtk->grid);

	grid_for_vtk->geometry_filter=vtkStructuredGridGeometryFilter::New();
	if (grid_for_vtk->geometry_filter==NULL) return NULL;
	grid_for_vtk->geometry_filter->SetExtent(0,input->GetNx(),0,input->GetNy(),0,input->GetNz());
	grid_for_vtk->geometry_filter->SetInput(grid_for_vtk->grid);

	grid_for_vtk->filter=vtkAppendPolyData::New();
	if (grid_for_vtk->filter==NULL) return NULL;
	grid_for_vtk->filter->AddInput(grid_for_vtk->geometry_filter->GetOutput());
	grid_for_vtk->filter->AddInput(grid_for_vtk->outline_filter->GetOutput());

	grid_for_vtk->mapper=vtkPolyDataMapper::New();
	if (grid_for_vtk->mapper==NULL) return NULL;
	grid_for_vtk->mapper->SetInput(grid_for_vtk->filter->GetOutput());

	grid_for_vtk->actor=vtkLODActor::New();
	if (grid_for_vtk->actor==NULL) return NULL;
	grid_for_vtk->actor->SetMapper(grid_for_vtk->mapper);
	grid_for_vtk->actor->GetProperty()->SetColor(1,0,1);

	Define(grid_for_vtk);

	return grid_for_vtk;
}
