/* capVTE

 Computational Advanced Propulsion Virtual Testing Environment
 (c) 2003 Lubos Brieda, cap lab, Vrginia Tech

 Created with assistance of: Dr. Joseph Wang, Raed Kafafy, Julien Pierru,
 Binh Tran and Ryan Stillwater.

*/

#include <iostream>
#include <iomanip>
using namespace std;

#include "vteVisualizationWindow.h"
#include "vteGlyphs.h"

#include <vtkPointData.h>
#include <vtkCamera.h>

#include <vtkStructuredPointsReader.h>
#include <vtkPiecewiseFunction.h>
#include <vtkColorTransferFunction.h>
#include <vtkVolumeProperty.h>
#include <vtkVolumeRayCastCompositeFunction.h>
#include <vtkVolumeRayCastMapper.h>
#include <vtkVolume.h>

vteVisualizationWindow::vteVisualizationWindow()
{
	show_glyphs=false;
	show_grid=false;
	show_title=false;
	show_geometry=false;
	show_scalar_gradient=false;
	show_isosurface=false;
	show_planes=false;
	show_scalar_bar=false;
	grid_component_index=0;
	max_frame=0;
	current_frame=0;

	renderer=NULL;
	//syscoordinate=NULL;
	qvtk_window=NULL;
	qvtk_widget=NULL;
	parent=NULL;
	input=NULL;

	title_for_vtk=NULL;
	scalar_bar_for_vtk=NULL;
	isosurface_for_vtk=NULL;
	cutting_plane_collection=NULL;
}

vteVisualizationWindow::~vteVisualizationWindow()
{
	if (input!=NULL) input->Delete();
}

/*COMBINE BELOW INTO CONSTRUCTOR*/
vteVisualizationWindow* vteVisualizationWindow::New(QWidget *parent,const char *file_name)
{
	vteVisualizationWindow *visualization_window=new vteVisualizationWindow;
	if (visualization_window==NULL) return NULL;

	visualization_window->input = vteFileReader::New(file_name);
	if (visualization_window->input==NULL) return NULL;

	visualization_window->parent=parent;
	return visualization_window;
}

void vteVisualizationWindow::Delete()
{
	delete this;
}


/******************************
vteVisualizationWindow::Execute()

Loads input and generates the render window
***********************************/
int vteVisualizationWindow::Execute()
{
	if (!input) return ErrorMessage("input not set");

	/*load the input file*/
	if (input->HasExecuted()==false)
	{
		input->SetStatusBar(status_bar);
		if (input->Execute()==vteERROR) return vteERROR;
	}

	//create renderer	
	renderer = vtkRenderer::New();
	if (renderer==NULL) return ErrorMessage("render is not defined");
	renderer->SetBackground(1,1,1);
	
	// VTK/Qt wedded
	qvtk_widget= new QVTKWidget();
	if (!qvtk_widget)
		return ErrorMessage("Creation of QVTK Widget failed!");

	qvtk_window = qvtk_widget->GetRenderWindow();
	if (qvtk_window==NULL) return ErrorMessage("window is not defined");
	
	/*combine QT and VTK*/
	qvtk_window->AddRenderer(renderer);
	
   	//create interactor
	//interactor = vtkRenderWindowInteractor::New();
	//if (interactor==NULL) return ErrorMessage("insufficient memory for interactor");
	//interactor->SetRenderWindow(window);
	
	/*get coordinate system
	syscoordinate=new vtkCoordinate();
	if (syscoordinate==NULL) return ErrorMessage("syscoordinate is not defined");
	syscoordinate->SetCoordinateSystemToNormalizedDisplay(); */

	/*make title*/
	title_for_vtk = vteTitleForVtk::New(input->GetTitle());
	ToggleTitle();

	/*is there geometry? if yes, show it*/
	if (HasGeometry())
	{
		ToggleGeometry();
	}

	/*get max frame*/
	if (HasGlyphs())
	{
		max_glyph_frame = new int [input->GetGlyphSetCollection()->GetNumberOfMembers()];
		for (int i=0;i<input->GetGlyphSetCollection()->GetNumberOfMembers();i++)
		{
		    max_glyph_frame[i] = input->GetGlyphSetCollection()->Get(i)->GetMaxFrame();
			if (max_glyph_frame[i]>max_frame) max_frame=max_glyph_frame[i];
		}

		ToggleGlyphs();
	}
	
	/*Grid*/
	if (HasGrid())
	{
		vteGridForVtk *grid_for_vtk=input->GetGridFrameCollection()->GetGridForVtk();

		max_grid_frame = input->GetGridFrameCollection()->GetMaxFrame();
		if (max_grid_frame>max_frame) max_frame=max_grid_frame;

		/*make scalar bar*/
		scalar_bar_for_vtk = vteScalarBarForVtk::New(grid_for_vtk);

		/*make isosurface*/
		isosurface_for_vtk = vteIsosurfaceForVtk::New(grid_for_vtk);

		/*maku cutting planes*/
		cutting_plane_collection = vteCuttingPlaneCollection::New();
		vteCuttingPlane *plane=cutting_plane_collection->NewCuttingPlane(grid_for_vtk);

		TogglePlanes();
		ToggleScalarBar();
	}


	/*little testing*/
	/*
	vtkStructuredPointsReader *reader = vtkStructuredPointsReader::New();
	reader->SetFileName ("ironProt.vtk");

	vtkPiecewiseFunction *opacityTransferFunction = vtkPiecewiseFunction::New();
	opacityTransferFunction->AddPoint (20,0.0);
	opacityTransferFunction->AddPoint (255,0.2);

	vtkColorTransferFunction *colorTransferFunction = vtkColorTransferFunction::New();
	colorTransferFunction->AddRGBPoint (0.0, 0.0, 0.0, 0.0);
	colorTransferFunction->AddRGBPoint (64.0, 1.0, 0.0, 0.0);
	colorTransferFunction->AddRGBPoint (128.0, 0.0, 0.0, 1.0);
	colorTransferFunction->AddRGBPoint (192.0, 0.0, 1.0, 0.0);
	colorTransferFunction->AddRGBPoint (255, 0.0, 0.2, 0.0);

	vtkVolumeProperty *volumeProperty = vtkVolumeProperty::New();
	volumeProperty->SetColor(colorTransferFunction);
	volumeProperty->SetScalarOpacity(opacityTransferFunction);
	volumeProperty->ShadeOn();
	volumeProperty->SetInterpolationTypeToLinear();

	vtkVolumeRayCastCompositeFunction *compositeFunction=vtkVolumeRayCastCompositeFunction::New();
	vtkVolumeRayCastMapper *volumeMapper = vtkVolumeRayCastMapper::New();
	volumeMapper->SetVolumeRayCastFunction(compositeFunction);
	volumeMapper->SetInputConnection (reader->GetOutputPort());
	
	vtkVolume *volume =vtkVolume::New();
	volume->SetMapper(volumeMapper);
	volume->SetProperty(volumeProperty);

	renderer->AddVolume(volume);*/


	renderer->ResetCamera();
	renderer->GetActiveCamera()->Elevation(-20);
	renderer->GetActiveCamera()->Azimuth(20);

	return vteOK;
}

//UpdateToFrame - updates all relevant information
//this should be moved to individual data types!
void vteVisualizationWindow::UpdateToFrame(long frame_index)
{
	long i,g;
	bool modified=false;

//	if (max_frame<1) return;
	current_frame=frame_index;
  
	if (current_frame>max_frame) current_frame=0;

	vteGlyphSetCollection *glyph_set_collection = input->GetGlyphSetCollection();
	vteGlyphSetForVtk *glyph_set_for_vtk=NULL;
	vteGlyphFrameCollection *glyph_frame_collection;
	vteGlyphFrame *glyph_frame;

	vteGlyph *glyph;
	int nog;
	unsigned long index_3, index_1;
	long glyph_frame_index,grid_frame_index;

	if ((show_glyphs==true) && HasGlyphs())
	{
		for (i=0;i<glyph_set_collection->GetNumberOfMembers();i++)
		{
			/*interpolate indes*/
			glyph_frame_index=(long)( (current_frame/(double)max_frame)*max_glyph_frame[i] );

			glyph_frame_collection=input->GetGlyphSetCollection()->Get(i);
			glyph_frame_collection->SetCurrentFrame(glyph_frame_index);
			glyph_frame=glyph_frame_collection->GetFrame();
			glyph=glyph_frame->GetGlyphs();
			nog=glyph_frame->GetNumberOfMembers();

			glyph_set_for_vtk = input->GetGlyphSetCollection()->Get(i)->GetGlyphSetForVtk();

			index_3=0;index_1=0;
			vtkStructuredGrid *grid = glyph_set_for_vtk->GetGrid();
			
			/*for (g=0;g<nog;g++)
			{
				ptr_to_nodes[index_3]=glyph[g].x;
				ptr_to_vectors[index_3]=glyph[g].u;
				index_3++;

				ptr_to_nodes[index_3]=glyph[g].y;
				ptr_to_vectors[index_3]=glyph[g].v;
				index_3++;
				ptr_to_nodes[index_3]=glyph[g].z;
				ptr_to_vectors[index_3]=glyph[g].w;
				index_3++;
				ptr_to_types[index_1]=glyph[g].t;
				index_1++;
			}*/
			vtkPoints *points = grid->GetPoints();
			vtkDataArray *normals = grid->GetPointData()->GetVectors();
			vtkDataArray *types = grid->GetPointData()->GetScalars();
	
			for (g=0;g<nog;g++)
			{
				points->SetPoint(g,glyph[g].x);
				normals->SetTuple(g,glyph[g].u);
				double t = glyph[g].t;
				types->SetTuple(g,&t);
			}
		
			glyph_set_for_vtk->GetFilter()->SetExtent(0,nog-1,0,1,0,1);
			glyph_set_for_vtk->GetGlyph()->Modified();
			glyph_set_for_vtk->GetFilter()->Modified();
			glyph_set_for_vtk->GetGrid()->Modified();
		}
		modified=true;
	}

	/*GRID*/
	if (HasGrid())
	{
		vteGridFrameCollection *grid_frame_collection = input->GetGridFrameCollection();
		vteGridForVtk *grid_for_vtk = grid_frame_collection->GetGridForVtk();
		
		grid_frame_index = (int) ( ((double)current_frame/(double)max_frame)*(double)max_grid_frame );
		grid_frame_collection->SetCurrentFrame(grid_frame_index);
		
		vteGridData *grid_data = grid_frame_collection->GetGridData();

		/*copy data*/
		memcpy(grid_for_vtk->GetPtrToScalars(),grid_data,grid_frame_collection->GetNumberOfNodes()*sizeof(vteGridData));

		if (HasIsosurface()) isosurface_for_vtk->GetFilter()->Modified();
		if (HasCuttingPlanes())
		{
     		for (i=0;i<cutting_plane_collection->GetNumberOfMembers();i++)
     		{
				cutting_plane_collection->GetCuttingPlane(i)->GetFilter()->Modified();
				cutting_plane_collection->GetCuttingPlane(i)->GetContourMapper()->Modified();
				cutting_plane_collection->GetCuttingPlane(i)->GetPlane()->Modified();
     		}
		}

		modified=true;
	}

	if (show_scalar_gradient==true)
	{
		vteGeometryCollection *geometry_collection=input->GetGeometryCollection();
		
		vteGeometryForVtk *geometry_for_vtk;
		for (i=0;i<geometry_collection->GetNumberOfMembers();i++)
		{
			geometry_for_vtk=geometry_collection->GetGeometry(i)->GetGeometryForVtk();
			ExtrapolateScalarGradient(geometry_for_vtk);
			geometry_for_vtk->GetMapper()->Modified();
		}
	}
    
	if (modified==true) qvtk_window->Render();
}

/**********************************
vteVisualizationWindow::Repaint

Forces redraw of onscreen image
***********************************/
void vteVisualizationWindow::Repaint()
{
	qvtk_widget->hide();
	qvtk_widget->show();		
}

/****************************
vteVisualizationWindow::ExtrapolateScalarGradient

Extrapolates mesh data onto geometry nodes
*****************************/
int vteVisualizationWindow::ExtrapolateScalarGradient(vteGeometryForVtk *geometry_for_vtk)
{
	long i,non;
	double f,g,A,B,C,D,E,F,G,H;
	double dx,dy,dz;
	double *scalars=geometry_for_vtk->GetScalars();
	vteGridForVtk *grid_for_vtk=input->GetGridFrameCollection()->GetGridForVtk();
	vteGridData *grid_data=grid_for_vtk->GetInput()->GetGridData();
	vteBoxIndex *box_index=geometry_for_vtk->GetBoxIndex();

	vteGridFrameCollection *grid_frame_collection=grid_for_vtk->GetInput();
	dx=grid_frame_collection->GetDx();
	dy=grid_frame_collection->GetDy();
	dz=grid_frame_collection->GetDz();

	non=geometry_for_vtk->GetInput()->GetNumberOfNodes();
	for (i=0;i<non;i++)
		{
		if (box_index[i].valid==true)
		 {
			A=grid_data[box_index[i].A];
			B=grid_data[box_index[i].B];
			C=grid_data[box_index[i].C];
			D=grid_data[box_index[i].D];
			E=grid_data[box_index[i].E];
			F=grid_data[box_index[i].F];
			G=grid_data[box_index[i].G];
			H=grid_data[box_index[i].H];

			f=A+dx*(B-A);
			f=f+dy*(C+dx*(D-C)-f);

			g=E+dx*(F-E);
			g=g+dy*(G+dx*(H-G)-g);

   			scalars[i]=f+dz*(g-f);
		}
		else  
			scalars[i]=grid_frame_collection->GetAbsoluteRange()->min;
    }

	return vteOK;
}

/****************************
vteVisualizatoinWindow::ToggleTitle

Turns title on or off
*****************************/
int vteVisualizationWindow::ToggleTitle()
{
	if (IsDefined(title_for_vtk))
	{
		if (show_title==true)
		{
			show_title=false;
			renderer->RemoveActor(title_for_vtk->GetActor());
		}
		else
		{
			show_title=true;
			renderer->AddActor(title_for_vtk->GetActor());
		}
	
		return vteOK;
	}
	
	return vteWARNING;
}

//toggle ScalarBar
int vteVisualizationWindow::ToggleScalarBar()
{
	if (IsDefined(scalar_bar_for_vtk))
	{
		if (show_scalar_bar==true)
		{
			show_scalar_bar=false;
			renderer->RemoveActor(scalar_bar_for_vtk->GetActor());
		}
		else
		{
			show_scalar_bar=true;
			renderer->AddActor(scalar_bar_for_vtk->GetActor());
		}
		return vteOK;
	}
	
	return vteWARNING;
}

//toggle geometry
int vteVisualizationWindow::ToggleGeometry()
{
	int i;

	if (HasGeometry())
	{
		if (show_geometry==true)
		{
			show_geometry=false;
			for (i=0;i<input->GetGeometryCollection()->GetNumberOfMembers();i++)
				renderer->RemoveActor(input->GetGeometryCollection()->GetGeometry(i)->GetGeometryForVtk()->GetActor());
		}
		else
		{
			show_geometry=true;
			for (i=0;i<input->GetGeometryCollection()->GetNumberOfMembers();i++)
				renderer->AddActor(input->GetGeometryCollection()->GetGeometry(i)->GetGeometryForVtk()->GetActor());
		}
	
		return vteOK;
	}
	
	return vteWARNING;
}

//toggle grid
int vteVisualizationWindow::ToggleGrid()
{
	if (HasGrid())
	{
		vteGridForVtk *grid_for_vtk=input->GetGridFrameCollection()->GetGridForVtk();
	
		if (show_grid==true)
		{
			show_grid=false;
			renderer->RemoveActor(grid_for_vtk->GetActor());
		}
		else
		{
			show_grid=true;
			renderer->AddActor(grid_for_vtk->GetActor());
		}
		
		return vteOK;
	}
	
	return vteWARNING;
}

//toggle particles
int vteVisualizationWindow::ToggleGlyphs()
{
	vteGlyphSetCollection *collection = input->GetGlyphSetCollection();
	
	int i;
	if (HasGlyphs())
	{
		if (show_glyphs==true)
		{
			show_glyphs=false;
			for (i=0;i<collection->GetNumberOfMembers();i++)
				renderer->RemoveActor(collection->Get(i)->GetGlyphSetForVtk()->GetActor());
		}
		else
		{
			show_glyphs=true;
			for (i=0;i<collection->GetNumberOfMembers();i++)
				renderer->AddActor(collection->Get(i)->GetGlyphSetForVtk()->GetActor());
		}
		return vteOK;
	}
	
	return vteWARNING;
}

//toggle ScalarGradient
int vteVisualizationWindow::ToggleScalarGradient()
{
	vteGridForVtk *grid_for_vtk=input->GetGridFrameCollection()->GetGridForVtk();
	vteGeometryCollection *geometry_collection=input->GetGeometryCollection();
	vteGeometryForVtk *geometry_for_vtk;
	double *scalars;
	double *color_scalars_backup;
	long i, non;

	if (HasGrid() && HasGeometry())
	{
		for (i=0;i<geometry_collection->GetNumberOfMembers();i++)
		{
			geometry_for_vtk=geometry_collection->GetGeometry(i)->GetGeometryForVtk();

			scalars=geometry_for_vtk->GetScalars();
			color_scalars_backup=geometry_for_vtk->GetColorScalarsBackup();
			non=geometry_for_vtk->GetInput()->GetNumberOfNodes();

  			if (show_scalar_gradient==true)
			{
				show_scalar_gradient=false;
				memcpy(scalars,color_scalars_backup,non*sizeof(float));
				geometry_for_vtk->GetMapper()->SetLookupTable(geometry_for_vtk->GetLut());
				geometry_for_vtk->GetMapper()->SetScalarRange(geometry_for_vtk->GetLut()->GetRange());
			}
			else
			{
				show_scalar_gradient=true;
				geometry_for_vtk->GetMapper()->SetLookupTable(grid_for_vtk->GetLut());
				geometry_for_vtk->GetMapper()->SetScalarRange(grid_for_vtk->GetUserRange()->min,grid_for_vtk->GetUserRange()->max);
				ExtrapolateScalarGradient(geometry_for_vtk);
			} //show ScalarGradient

			geometry_for_vtk->GetMapper()->Modified();
		} //for members
		return vteOK;
	} 
	
	return vteWARNING;
}

//toggle planes
int vteVisualizationWindow::TogglePlanes()
{
	int i;
	if (HasCuttingPlanes())
	{
		if (show_planes==true)
		{
			show_planes=false;
			for (i=0;i<cutting_plane_collection->GetNumberOfMembers();i++)
			{
				renderer->RemoveActor(cutting_plane_collection->GetCuttingPlane(i)->GetContourActor());
				renderer->RemoveActor(cutting_plane_collection->GetCuttingPlane(i)->GetCutterActor());
			}  
		}
		else
		{
			show_planes=true;
			for (i=0;i<cutting_plane_collection->GetNumberOfMembers();i++)
			{
				renderer->AddActor(cutting_plane_collection->GetCuttingPlane(i)->GetContourActor());
				renderer->AddActor(cutting_plane_collection->GetCuttingPlane(i)->GetCutterActor());
			}  
		}
	
		return vteOK;
	}

	return vteWARNING;
}

//toggle isosurface
int vteVisualizationWindow::ToggleIsosurface()
{
	if (HasIsosurface())
	{
		if (show_isosurface==true)
		{
			show_isosurface=false;
			renderer->RemoveActor(isosurface_for_vtk->GetActor());
		}
		else
		{
			show_isosurface=true;
			renderer->AddActor(isosurface_for_vtk->GetActor());
		}
		return vteOK;
	}
	
	return vteWARNING;
}

int vteVisualizationWindow::WriteJPEG(char *file_name,int quality)
{
	vtkWindowToImageFilter *w2i_filter;
	vtkJPEGWriter *jpeg_writer;

	w2i_filter=vtkWindowToImageFilter::New();
	w2i_filter->SetInput(qvtk_window);

	jpeg_writer=vtkJPEGWriter::New();
	jpeg_writer->SetInput(w2i_filter->GetOutput());
	jpeg_writer->SetQuality(quality);
	jpeg_writer->SetFileName(file_name);
	jpeg_writer->ProgressiveOff();
	jpeg_writer->Write();

	jpeg_writer->Delete();
	w2i_filter->Delete();
	return vteOK;
}

int vteVisualizationWindow::WriteIV(char *file_name)
{
	vtkIVExporter *iv_exporter;

	iv_exporter=vtkIVExporter::New();
	iv_exporter->SetInput(qvtk_window);
	iv_exporter->SetFileName(file_name);

	iv_exporter->Write();
	iv_exporter->Delete();
	return vteOK;
}

int vteVisualizationWindow::WriteVRML(char *file_name)
{
	vtkVRMLExporter *vrml_exporter;

	vrml_exporter=vtkVRMLExporter::New();
	vrml_exporter->SetInput(qvtk_window);
	vrml_exporter->SetFileName(file_name);

	vrml_exporter->Write();
	vrml_exporter->Delete();
	return vteOK;
}
